#!/devtool/local/bin/perl
#
#    LCAEnv.pl - Creates and syncs a complete LCA environment.
#
#    @(#)LCAEnv     2003-02-18
#
#    U. Jansen, SAP AG
#
#    ========== licence begin LGPL
#    Copyright (C) 2002 SAP AG
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    ========== licence end
#

##############################################################################
# Static includes:

use strict;
use Getopt::Long;
use File::Path;
use File::Copy;
use APOSettings;

##############################################################################
# Console Preparation (redirect STDERR to STDOUT and disable buffering):

close   STDERR;
open    STDERR, ">&STDOUT";
$| = 1;

##############################################################################
# Global variables:

my $ProgramName         = "LCAEnv";
my $Version             = "1";
my $Revision            = "00";
my $Date                = "2003-03-13";

my %Opts                = ();
my $Glob                = {};
$Glob->{'P4Port'}       = "perforce3011.wdf.sap.corp:3011";
$Glob->{'PathSep'}      = ($^O =~ /MSWin32/i) ? "\\" : "/";
$ENV{'P4PORT'}          = $Glob->{'P4Port'};

my $DoCreate            = 0;
my $DoMake              = 0;
my $DoClear             = 0;

##############################################################################
# Main routine:
##############################################################################

unless ($ENV{'TOOL'})
{
    print "%TOOL% is not set in the environment!\n";
    exit (10);
}

$ProgramName = $2 if ($0 =~ /^(.*[\/\\])?([^\/\\]+)$/);
$ProgramName =~ s/\.pl//g;

print "\n$ProgramName v$Version.$Revision ($Date), (c)2003 SAP AG\n";

if ($ProgramName eq "LCAMakeEnv")
{
    $DoCreate   = 1;

    if (!GetOptions( \%Opts, 'h|help|?', 'own=s', 'rel=s', 'cl=i', 'lc=s', 'p4u=s', 'p4c=s', 'p4r=s', 'fr', 'v', 'y', 'm=i', 'M=s', 'cfg=s', 'po', 'prod', '_dosbatch') || $Opts{'h'} || @ARGV)
    {
        Usage();
    }
}
elsif ($ProgramName eq "LCAClearEnv")
{
    $DoClear    = 1;

    if (!GetOptions( \%Opts, 'h|help|?', 'own=s', 'cfg=s', 'v', 'y') || $Opts{'h'} || @ARGV)
    {
        Usage();
    }
}
else
{
    $DoCreate   = 1;

    if (!GetOptions( \%Opts, 'h|help|?', 'own=s', 'rel=s', 'cl=i', 'lc=s', 'p4u=s', 'p4c=s', 'p4r=s', 'fr', 'v', 'y', 'm=i', 'M=s', 'c', 'cfg=s', 'po', 'prod', '_dosbatch') || $Opts{'h'} || @ARGV)
    {
        Usage();
    }
}

if ($Opts{'cfg'})
{
    $Glob = LoadCfg($Opts{'cfg'});
}

if ($DoCreate)
{
    $Glob->{'APORel'}   = (defined $Opts{'rel'}  ? $Opts{'rel'}  : ( defined $Glob->{'APORel'}   ? $Glob->{'APORel'}   : AskAPORel() ));
    $Glob->{'ChList'}   = (defined $Opts{'lc'}   ? $Opts{'lc'}   : ( defined $Glob->{'ChList'}   ? $Glob->{'ChList'}   : AskChList() ));
    $Glob->{'LCVer'}    = (defined $Opts{'lc'}   ? $Opts{'lc'}   : ( defined $Glob->{'LCVer'}    ? $Glob->{'LCVer'}    : AskLCVer() ));
    $Glob->{'P4User'}   = (defined $Opts{'p4u'}  ? $Opts{'p4u'}  : ( defined $Glob->{'P4User'}   ? $Glob->{'P4User'}   : AskP4User() ));
    $Glob->{'P4Client'} = (defined $Opts{'p4c'}  ? $Opts{'p4c'}  : ( defined $Glob->{'P4Client'} ? $Glob->{'P4Client'} : AskP4Client() ));
    $Glob->{'P4Root'}   = (defined $Opts{'p4r'}  ? $Opts{'p4r'}  : ( defined $Glob->{'P4Root'}   ? $Glob->{'P4Root'}   : AskP4Root() ));
    $Glob->{'Own'}      = (defined $Opts{'own'}  ? $Opts{'own'}  : ( defined $Glob->{'Own'}      ? $Glob->{'Own'}      : AskOwn() ));
    $Glob->{'Make'}     = (defined $Opts{'M'}    ? $Opts{'M'}    : ( defined $Opts{'m'}          ? $Opts{'m'}          : ( defined $Glob->{'Make'} ? $Glob->{'Make'} : AskMake() )));
    $Glob->{'RELSTAT'}  = (defined $Opts{'prod'} ? 1 : 0);

    $DoClear = (defined $Opts{'c'} ? 1 : 0);

    print "\n\nThe following settings will be used:\n\n";
    print "APO release          : $Glob->{'APORel'}\n";
    print "Change list          : " . ($Glob->{'ChList'} > 0 ? $Glob->{'ChList'} : "<latest>") . "\n";
    print "LC version           : ". ($Glob->{'LCVer'} ? $Glob->{'LCVer'} : "<default>") . "\n";
    print "\$Own directory       : $Glob->{'Own'}\n";
    print "Perforce server      : $Glob->{'P4Port'}\n";
    print "Perforce user        : $Glob->{'P4User'}\n";
    print "Perforce client      : $Glob->{'P4Client'}\n";
    print "Perforce client root : $Glob->{'P4Root'}\n";
    print "Make                 : " . ($Glob->{'Make'} eq "0" ? "Nothing" : ($Glob->{'Make'} eq "1" ? "Default" : $Glob->{'Make'})) . "\n";
    print "RELSTAT              : PROD\n" if ($Glob->{'RELSTAT'} == 1);

    unless($Opts{'y'} or (RequestInput("\nIs this correct? [Y/n] : ", "y", "[yYnN]", "Input incorrect! Please, use only [y or n].\n") =~ /^[1jy]/i))
    {
        print "\nAbort!\n";
        exit(5);
    }

    SetOwn();
    SaveCfg("$Glob->{'Own'}$Glob->{'PathSep'}LCAEnv.cfg", $Glob);

    if (defined $Opts{'po'})
    {
        print "\nPreparation finished. Exiting...\n";
        exit(0);
    }

    SetAPORel();
    SetP4User();
    SetP4Client();
    SetP4Root();
    SetP4ClientView();
    SetChList();
    P4Sync();
    MakeEnv();

    if ($Glob->{'Make'} ne "0")
    {
        DoMake();
    }
}

if ($DoClear)
{
    $Glob->{'Own'} = (defined $Opts{'own'}  ? $Opts{'own'}  : ($Glob->{'Own'} ? $Glob->{'Own'} : $ENV{'OWN'}));

    unless (defined $Glob->{'Own'})
    {
        print "\nError! \$Own is not set! Exiting...\n";
        exit(10);
    }

    unless ($Opts{'y'} or (RequestInput("\nDeleting all entries for '$Glob->{'Own'}'.\nIs this correct? [y/N] : ", "n", "[yYnN]", "Input incorrect. Please, just use [y or n].\n") =~ /^[1jy]/i))
    {
        print "\nClean-up cancelled. Use '-own' or '-cfg' to specify the release to delete.\n\n";
        exit(5);
    }
    $Glob = LoadCfg("$Glob->{'Own'}/LCAEnv.cfg");

    ClearAll();
}

exit(0);

##############################################################################
# AskAPORel() - Asks for the APO release to use
##############################################################################

sub AskAPORel
{
    my $count = 1;
    my @apokeys;
    my $erg;
    print "\nAvailable APO releases:\n\n";
    foreach (reverse sort keys %APOSettings::Depots)
    {
        my $outstr = sprintf("[%2u] = %s\n", $count, $_);
        print "$outstr";
        $apokeys[$count++] = $_;
    }

    print "\nPlease enter the APO release [1 - " . ($count - 1 ). "] : ";
    $erg = <STDIN>;
    chomp $erg;
    while (($erg <= 0) or ($erg >= $count))
    {
        print "Input incorrect. Please use numbers between 1 - " . ($count - 1) . ".\n";
        print "\nPlease enter the APO release [1 - " . ($count - 1 ). "] : ";
        $erg = <STDIN>;
        chomp $erg;
    }

    return $apokeys[$erg];
}

##############################################################################
# SetAPORel() - Sets the APO release to use
##############################################################################

sub SetAPORel
{
    unless ($APOSettings::Depots{$Glob->{'APORel'}})
    {
        print "\nError! The APO Release you entered ($Glob->{'APORel'}) is unknown!\nExiting...\n";
        exit (15);
    }
    return;
}

##############################################################################
# AskChList() - Asks for the ChangeList which will be used to sync the sources
##############################################################################

sub AskChList
{
    if ($Opts{'y'})
    {
        return 0;
    }

    my $res;
    print "\nPlease enter the change list you want to use.\n\nChangeList ['0' or <Enter> = latest] : ";
    $res = <STDIN>;
    chomp $res;
    if ($res =~ /^$/)
    {
        return 0;
    }
    while ($res !~ /^\d+$/)
    {
        print "\nInput incorrect. Please, use only numbers or <Enter>.\n\nChangeList ['0' or <Enter> = latest] : ";
        $res = <STDIN>;
        chomp $res;
    }

    return $res;
}

##############################################################################
# AskLCVer() - Asks for the LC-Version which will be used
##############################################################################

sub AskLCVer
{
    if ($Opts{'y'})
    {
        return 0;
    }

    my $res;
    print "\nPlease enter the LC Version you want to use.\n\nLC Version ['0' or <Enter> = default version] : ";
    $res = <STDIN>;
    chomp $res;
    if ($res =~ /^$/)
    {
        return 0;
    }
    while ($res !~ /^\S+$/)
    {
        print "\nInput incorrect. Please, type in a single word or <Enter>.\n\nChangeList ['0' or <Enter> = default version] : ";
        $res = <STDIN>;
        chomp $res;
    }

    return $res;
}

##############################################################################
# AskOwn() - Determines the "Own" directory to use.
##############################################################################

sub AskOwn
{
    my $curdir      = GetCurrentDir();
    my $reldef      = $Glob->{'P4Root'} . $Glob->{'PathSep'} . $Glob->{'APORel'};

    print "\nPlease select which \$Own directory you want to use:\n\n";

    print "[c] = use the current directory ($curdir)\n";
    print "[p] = use P4 client root ($Glob->{'P4Root'})\n";
    print "[r] = use releases default directory ($reldef), default\n";
    print "[s] = select a different client name manually\n\n";

    my $sel = RequestInput("Please select: ", "r", "[cCpPrRsS]", "Input incorrect. Please, just use [c, p, r or s].\n");

    if ($sel =~ /^c$/i)
    {

        return $curdir;
    }
    elsif ($sel =~ /^p$/i)
    {
        return $Glob->{'P4Root'};
    }
    elsif ($sel =~ /^r$/i)
    {
        return $reldef;
    }
    else
    {
        return ($Glob->{'P4Root'} . $Glob->{'PathSep'} . RequestInput("\$Glob->{'Own'} directory: $Glob->{'P4Root'}$Glob->{'PathSep'}", "", ".+", "Input incorrect. Please, just use normal characters.\n"));
    }
}

##############################################################################
# SetOwn() - Sets the "Own" directory to use.
##############################################################################

sub SetOwn
{
    unless ($Glob->{'Own'} =~ /^\Q$Glob->{'P4Root'}\E/i)
    {
        print "\nError! The \$Glob->{'Own'} directory ($Glob->{'Own'}) has to be\na sub directory of the Perforce client root you want to use!\nExiting...\n";
        exit (15);
    }

    unless(-d $Glob->{'Own'})
    {
        if($Opts{'y'} or (RequestInput("\nDirectory '$Glob->{'Own'}' does not exist. Create ? [Y/n] : ", "y", "[yYnN]", "Input incorrect. Please use [y or n].\n") =~ /^[1jy]/i))
        {
            print "\nCreating directory '$Glob->{'Own'}'...\n";
            File::Path::mkpath(($Glob->{'Own'}), 0, 0777) or die "Can't create directory '$Glob->{'Own'}'!\n$!\n";
            $ENV{'OWN'} = $Glob->{'Own'};
        }
        else
        {
            print "Can't continue without \$Glob->{'Own'}. Exiting...\n";
            exit (10);
        }
    }
}

##############################################################################
# AskMake() - Asks wether to make (and what) or not
##############################################################################

sub AskMake
{
    my $target;
    my $sel;

    print "\nPlease select a make action you want to perform:\n\n";
    print "[n] = no make, default\n";
    print "[d] = default make (lcapps.mac)\n";
    print "[s] = enter a different target to make, manually\n\n";

    if ($Opts{'y'})
    {
        $sel = "n";
    }
    else
    {
        $sel = RequestInput("Please select: ", "n", "[nNdDsS]", "Input incorrect. Please, just use [n, d or s].\n");
    }

    if ($sel =~ /^n$/i)
    {
        $target = "0";
    }
    elsif ($sel =~ /^d$/i)
    {
        $target = "1";
    }
    elsif ($sel =~ /^s$/i)
    {
        $target = RequestInput("Make target: ", "", ".+", "Input incorrect. Please, enter a single target name.\n");
    }

    return $target;
}

##############################################################################
# AskP4User() - Asks for the perforce user
##############################################################################

sub AskP4User
{
    my $login = getlogin;
    my $p4user;

    print "\nPlease select which Perforce user you want to use:\n\n";
    print "[l] = use your login name ($login), default\n";

    if ($ENV{'P4USER'})
    {
        print "[p] = use current Perforce user ($ENV{'P4USER'})\n";
    }

    print "[s] = select a different user name manually\n\n";

    my $sel = RequestInput("Please select: ", "l", $ENV{'P4USER'} ? "[lLpPsS]" : "[lLsS]", "Input incorrect. Please, just use " . ($ENV{'P4USER'} ? "[l, p or s].\n" : "[l or s].\n"));

    if ($sel =~ /^l$/i)
    {
        $p4user = $login;
    }
    elsif ($sel =~ /^p$/i)
    {
        $p4user = $ENV{'P4USER'};
    }
    else
    {
        $p4user = RequestInput("Perforce user: ", "", "[a-zA-Z0-9_]+", "Input incorrect. Please, just use [a-z, A-Z, 0-9 or _].\n");
    }

    $ENV{'P4USER'} = $p4user;

    return $p4user;
}


##############################################################################
# SetP4User() - Sets the Perforce user
##############################################################################

sub SetP4User
{
    my $found = 0;

    open USER_OUT, "p4 -p $Glob->{'P4Port'} user -o $Glob->{'P4User'} |" or die "Can't open 'p4 user -o $Glob->{'P4User'}'!\n$!\n";
    while (<USER_OUT>)
    {
        $found = 1 if (/^Access:/);
    }
    close USER_OUT or die "Can't close 'p4 user -i $Glob->{'P4User'}'!\n$!\n";

    if ($found)
    {
        print "\nUsing existing Perforce user '$Glob->{'P4User'}'...\n";
    }
    else
    {
        print "\nPerforce user '$Glob->{'P4User'}' doesn't exist. Create? [y/N] : ";
        my $res = <STDIN>;
        unless ($res =~ /^[yY]/)
        {
            print "\nUser not created! Aborting...";
            exit(5);
        }

        my ($fullname, $email) = ($1, $2) if (qx($ENV{TOOL}/bin/tel.pl $Glob->{'P4User'}) =~ /^\"([\w|\s]+)\".*,\s(\S+@\S+),/);

        unless ($fullname)
        {
            print "\nCan't determine the full name for user $Glob->{'P4User'}.\n";
            $fullname = RequestInput("Full name: ", "", "[a-zA-Z0-9_]+", "Input incorrect. Please, just use [a-z, A-Z, 0-9 or _].\n");
        }

        unless ($email)
        {
            print "\nCan't determine email address for user $Glob->{'P4User'}.\n";
            $email = RequestInput("Email address: ", "", "[a-zA-Z0-9_@]+", "Input incorrect. Please, just use [a-z, A-Z, 0-9, @ or _].\n");
        }

        print "Creating Perforce user:\n\nUser     = $Glob->{'P4User'}\nFullname = $fullname\nEmail    = $email\n\n";

        open USER_IN, "| p4 -p $Glob->{'P4Port'} user -i" or die "Can't open 'p4 user -i'!\n$!\n";
        print USER_IN "User: $Glob->{'P4User'}\n";
        print USER_IN "Email: $email\n";
        print USER_IN "FullName: $fullname\n";
        close USER_IN or die "Can't 'close p4 user -i'!\n$!\n";

    }
}

##############################################################################
# AskP4Client() - Asks for the Perforce client
##############################################################################

sub AskP4Client
{
    my $hostname = qx(hostname);
    my $p4client;

    chomp $hostname;

    print "\nPlease select which Perforce client you want to use:\n\n";
    print "[h] = use the local host name ($hostname), default\n";

    if ($ENV{'P4CLIENT'})
    {
        print "[p] = use current Perforce client ($ENV{'P4CLIENT'})\n";
    }

    print "[s] = select a different client name manually\n\n";

    my $sel = RequestInput("Please select: ", "h", $ENV{'P4CLIENT'} ? "[hHpPsS]" : "[hHsS]", "Input incorrect. Please, just use " . ($ENV{'P4CLIENT'} ? "[h, p or s].\n" : "[h or s].\n"));

    if ($sel =~ /^h$/i)
    {
        $p4client = $hostname;
    }
    elsif ($sel =~ /^p$/i)
    {
        $p4client = $ENV{'P4CLIENT'};
    }
    else
    {
        $p4client = RequestInput("Perforce client: ", "", "[a-zA-Z0-9_]+", "Input incorrect. Please, just use [a-z, A-Z, 0-9 or _].\n");
    }

    $ENV{'P4CLIENT'} = $p4client;

    return $p4client;
}


##############################################################################
# SetP4Client() - Sets the Perforce client
##############################################################################

sub SetP4Client
{
    my $found = 0;
    my $hostname = qx(hostname);

    chomp $hostname;

    open CLIENT_OUT, "p4 clients |" or die "Can't open 'p4 clients'!\n$!\n";
    while (<CLIENT_OUT>)
    {
        $found = 1 if (/^Client\s$Glob->{'P4Client'}\s.*/);
    }
    close CLIENT_OUT or die "Can't close 'p4 clients'!\n$!\n";

    if ($found)
    {
        print "\nUsing existing Perforce client '$Glob->{'P4Client'}'...\n";
    }
    else
    {
        unless ($Opts{'y'} or (RequestInput("\nPerforce client '$Glob->{'P4Client'}' doesn't exist. Create? [Y/N] : ", "y", "[yYnN]", "Input incorrect. Please, use [y or n].\n") =~ /^[1jy]/i))
        {
            print "\nClient not created! Aborting...";
            exit(5);
        }

        print "\nCreating client '$Glob->{'P4Client'}'...\n";
        open( CLIENT_IN, "| p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -i") or die "open p4 client -i failed! \n$!\n";
        print CLIENT_IN "Client: $Glob->{'P4Client'}\n";
        print CLIENT_IN "Owner: $Glob->{'P4User'}\n";
        print CLIENT_IN "Host: $hostname\n";
        print CLIENT_IN "Description:\n\tCreated by LCAEnv.\n";
        print CLIENT_IN "Root:\t$Glob->{'P4Root'}\n";
        print CLIENT_IN "Options:\tnoallwrite noclobber nocompress crlf unlocked nomodtime normdir\n";
        print CLIENT_IN "View:\n";
        close CLIENT_IN or die "Can't save the p4 client! \n$!\n";
    }
}

##############################################################################
# AskP4Root() - Asks for the Perforce client root to use
##############################################################################

sub AskP4Root
{
    my $clientroot  = GetP4ClientRoot();
    my $curdir      = GetCurrentDir();
    my $p4root;

    print "\nPlease select which client root you want to use for your P4 client:\n\n";

    print "[c] = use the current directory ($curdir)\n";

    if ($clientroot)
    {
        print "[p] = use current P4 client root ($clientroot)\n";
    }

    print "[s] = select a different client name manually\n\n";

    my $sel = RequestInput("Please select: ", "", $clientroot ? "[cCpPsS]" : "[cCsS]", "Input incorrect. Please, just use " . ($clientroot ? "[c, p or s].\n" : "[o or s].\n"));

    if ($sel =~ /^c$/i)
    {

        $p4root = $curdir;
    }
    elsif ($sel =~ /^p$/i)
    {
        $p4root = $clientroot;
    }
    else
    {
        $p4root = RequestInput("Perforce client root: ", "", ".+", "Input incorrect. Please, just use normal characters.\n");
    }

    return $p4root;

}

##############################################################################
# SetP4Root() - Sets the new P4Client Root
##############################################################################

sub SetP4Root
{
    my $clientroot = GetP4ClientRoot();

    if (lc($clientroot) ne lc($Glob->{'P4Root'}))
    {
        unless ($Opts{'y'} or (RequestInput("\nThe Perforce client root you entered differs from actual value:\nOld root = $clientroot\nNew root = $Glob->{'P4Root'}\n\nDo you want to change it to the new value ? [y/N] : ", "n", "[yYnN]", "Input incorrect. Please, use only [y or n].\n") =~ /^[1jy]/i))
        {
            print "\nPerforce client root not changed. Aborting...\n";
            exit(5);
        }

        open CLIENT_IN, "| p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -i"            or die "Can't open p4 client -i!\n$!\n";
        open CLIENT_OUT, "p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -o $Glob->{'P4Client'} |" or die "Can't open p4 client -o $Glob->{'P4Client'}!\n$!\n";
        while (<CLIENT_OUT>)
        {
            if (/^Root:\s+/)
            {
                print CLIENT_IN "Root:\t$Glob->{'P4Root'}\n";
            }
            else
            {
                print CLIENT_IN $_;
            }
        }
        close CLIENT_OUT or die "Can't close p4 client -o $Glob->{'P4Client'}!\n$!\n";
        close CLIENT_IN  or die "Can't close p4 client -i!\n$!\n";
        print "\nPerforce client root set to $Glob->{'P4Root'}...\n";
    }
    else
    {
        print "\nPerforce client root is up-to-date...\n";
    }
}

##############################################################################
# SetP4ClientView() - Adds a new client to Perforce
##############################################################################

sub SetP4ClientView
{
    my %Views;
    my $actview = $APOSettings::Depots{$Glob->{'APORel'}} . "/" . $Glob->{'APORel'};
    my $actpath = $Glob->{'Own'} . "/sys/src";
    my $newwritten = 0;
    my $viewfound = 0;

    $actpath =~ s/\Q$Glob->{'P4Root'}\E//i;
    $actpath =~ tr/\\/\//;
    $actpath = "//$Glob->{'P4Client'}" . $actpath;

    open CLIENT_OUT, "p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -o $Glob->{'P4Client'} |" or die "Can't open p4 client -o $Glob->{'P4Client'}!\n$!\n";
    while (<CLIENT_OUT>) { $Views{$1} = $2 if (/^\s+(\S+)\/\.\.\.\s+(\S+)\/\.\.\..*/) }
    close CLIENT_OUT or die "Can't close p4 client -o $Glob->{'P4Client'}!\n$!\n";

    if (defined $Views{$actview})
    {
        if (lc($Views{$actview}) eq lc($actpath))
        {
            print "\nPerforce client view path is up-to-date...\n";
            return;
        }
        unless ($Opts{'y'} or (RequestInput("\nThe Perforce client view path you entered differs from actual value:\nOld view path = $Views{$actview}\nNew view path = $actpath\n\nDo you want to change it to the new value ? [y/N] : ", "n", "[yYnN]", "Input incorrect. Please, use only [y or n].\n") =~ /^[1jy]/i))
        {
            print "\nPerforce client view for Rel $Glob->{'APORel'} has not been changed.\nAborting...\n";
            exit(5);
        }
    }

    print "\nUpdating Perforce client view path...\n";

    open CLIENT_IN, "| p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -i"            or die "Can't open p4 client -i!\n$!\n";
    open CLIENT_OUT, "p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -o $Glob->{'P4Client'} |" or die "Can't open p4 client -o $Glob->{'P4Client'}!\n$!\n";
    while (<CLIENT_OUT>)
    {
        print CLIENT_IN $_;
        if (/^View:.*/)
        {
            $viewfound = 1;
            last;
        }
    }

    unless ($viewfound)
    {
        print CLIENT_IN "View:\n";
    }

    foreach my $key (keys %Views)
    {
        if ($key eq $actview)
        {
            print CLIENT_IN "\t$actview/... $actpath/...\n";
            $newwritten = 1;
        }
        else
        {
            print CLIENT_IN "\t$key/... $Views{$key}/...\n";
        }
    }

    unless ($newwritten)
    {
        print CLIENT_IN "\t$actview/... $actpath/...\n";
    }

    close CLIENT_OUT or die "Can't close p4 client -o $Glob->{'P4Client'}!\n$!\n";
    close CLIENT_IN  or die "Can't close p4 client -i!\n$!\n";
}

##############################################################################
# RemP4ClientView () - Removes a specific client view and ev. the whole client
##############################################################################

sub RemP4ClientView
{
    my $actview = $APOSettings::Depots{$Glob->{'APORel'}} . "/" . $Glob->{'APORel'};
    my $views = 0;

    print "\nRemoving client entries...\n";

    open CLIENT_IN, "| p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -i"            or die "Can't open p4 client -i!\n$!\n";
    open CLIENT_OUT, "p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -o $Glob->{'P4Client'} |" or die "Can't open p4 client -o $Glob->{'P4Client'}!\n$!\n";
    while (<CLIENT_OUT>)
    {
        $views ++ if (/^\s+(\S+)\/\.\.\.\s+(\S+)\/\.\.\./);
        print CLIENT_IN $_ unless (/^\s+$actview\/\.\.\..*/);
    }

    close CLIENT_OUT or die "Can't close p4 client -o $Glob->{'P4Client'}!\n$!\n";
    close CLIENT_IN  or die "Can't close p4 client -i!\n$!\n";

    if ($views == 1)
    {
        if ($Opts{'y'} or (RequestInput("\nThe actual client ($Glob->{'P4Client'}) has no more views.\nRemove the client? [y/N] : ", "n", "[yYnN]", "Input incorrect. Please, use just [y or n].\n") =~ /^[1jy]/i))
        {
            my $rc = system("p4 -p $Glob->{'P4Port'} client -d $Glob->{'P4Client'}");

            if ($rc)
            {
                print "\nError! Client '$Glob->{'P4Client'}' could not be deleted! Continuing...\n";
            }
        }
        else
        {
            print "\nClient '$Glob->{'P4Client'}' has not been deleted...\n";
        }
    }

}

##############################################################################
# SetChList() - Determines the correct changelist number
##############################################################################

sub SetChList
{
    my $branch = $APOSettings::Depots{$Glob->{'APORel'}} . "/" . $Glob->{'APORel'} . "/...";

    unless ($Glob->{'ChList'})
    {
        open CHLIST_OUT, "p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} changes -m 1 -s submitted $branch |" or die "Can't open p4 changes -m 1 -s submitted $branch\n$!\n";
        while(<CHLIST_OUT>)
        {
            if (/^Change\s*(\d+)\s/)
            {
                $Glob->{'ChList'} = $1;
            }
        }
        close CHLIST_OUT or die "Can't close p4 changes -m 1 -s submitted $branch!\n$!\n";
    }

    unless ($Glob->{'ChList'})
    {
        print "\nCouldn't determine change list! Please make sure that the Perforce user\nyou entered has access rights for this release ($Glob->{'APORel'}).\nExiting...\n";
        exit (15);
    }

    # Additionally, log the changelist number into a file in $Own for the make progcess:
    open CHLIST_OUT, ">$Glob->{'Own'}$Glob->{'PathSep'}changelist.log" or die "Can't open '$Glob->{'Own'}$Glob->{'PathSep'}changelist.log' for writing!\n$!\n";
    print CHLIST_OUT "$Glob->{'ChList'}\n";
    close CHLIST_OUT or die "Can't close '$Glob->{'Own'}$Glob->{'PathSep'}changelist.log'!\n$!\n";

    print "\nRelease change list is $Glob->{'ChList'}...\n";
}

##############################################################################
# P4Sync() - Sync the actual source tree
##############################################################################

sub P4Sync
{
    my $branch      = $APOSettings::Depots{$Glob->{'APORel'}} . "/" . $Glob->{'APORel'} . "/...";
    my $cmdparam    = ($Glob->{'ChList'} ne "none" ? "$branch\@$Glob->{'ChList'}" : "$branch/...#none");
    my $rc          = 0;

    if ($Opts{'fr'})
    {
        $cmdparam = "-f " . $cmdparam;
        print "\nDoing a complete re-sync of [$Glob->{'APORel'} \@ $Glob->{'ChList'}]" . ($Opts{'v'} ? " :\n\n" : "...\n");
    }
    else
    {
        if ($Glob->{'ChList'} ne "none")
        {
            print "\nSynchronizing files [$Glob->{'APORel'} \@ $Glob->{'ChList'}]" . ($Opts{'v'} ? " :\n\n" : "...\n");
        }
        else
        {
            print "\nRemoving all files from [$Glob->{'APORel'}]" . ($Opts{'v'} ? " :\n\n" : "...\n");
        }
    }

    $| = 1;

    open CMD_OUT, "p4 -s -p $Glob->{'P4Port'} -u $Glob->{'P4User'} -c $Glob->{'P4Client'} sync $cmdparam |" or die "Can't open p4 sync $cmdparam!\n$!\n";
    while (<CMD_OUT>)
    {
        if    (/^(.*):\s(.*)\s-\sadded/)       { print "Add     : $2\n" if ($Opts{'v'}) }
        elsif (/^(.*):\s(.*)\s-\srefreshing/)  { print "Refresh : $2\n" if ($Opts{'v'}) }
        elsif (/^(.*):\s(.*)\s-\supdating/)    { print "Update  : $2\n" if ($Opts{'v'}) }
        elsif (/^(.*):\s(.*)\s-\sdeleted/)     { print "Deleted : $2\n" if ($Opts{'v'}) }
        elsif (/^(.*):.*up-to-date/)           { print "Info    : All files are up-to-date.\n"    }
        elsif (/^error:\s(.*)/)                { print "Error   : $1\n"    }
        elsif (/^info:\s(.*)/)                 { print "Info    : $1\n" if ($Opts{'v'}) }
        elsif (/^exit:\s(\d+)/)                { $rc = $1;                 }
        else                                   { print "$_" if ($Opts{'v'}) }
    }
    close CMD_OUT or $rc = 40;

    if ($Glob->{'ChList'} ne "none")
    {
        print ($rc ? "\nSync failed.\n" : "\nSync OK.\n");
    }
    else
    {
        print ($rc ? "\nFile removal failed.\n" : "\nFile removal OK.\n");
    }
}

##############################################################################
# GetP4ClientRoot() - Retrieves the root directory for a P4 client
##############################################################################

sub GetP4ClientRoot
{
    my $rootpath;
    my $found = 0;

    open CLIENT_OUT, "p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} clients |" or die "Can't open 'p4 clients'!\n$!\n";
    while (<CLIENT_OUT>)
    {
        $found = 1 if (/^Client\s$Glob->{'P4Client'}\s.*/);
    }
    close CLIENT_OUT or die "Can't close 'p4 clients'!\n$!\n";

    return "" unless ($found);

    open CLIENT_OUT, "p4 -p $Glob->{'P4Port'} -u $Glob->{'P4User'} client -o $Glob->{'P4Client'} |" or die "Can't open p4 client -o $Glob->{'P4Client'}! \n$!\n";
    while (<CLIENT_OUT>)
    {
        $rootpath = $1 if (/^Root:\s+(\S+)/);
    }
    close CLIENT_OUT or die "Can't close p4 client -o $Glob->{'P4Client'}! \n$!\n";

    return $rootpath;
}

##############################################################################
# MakeEnv() - Sets the current environment.
##############################################################################

sub MakeEnv
{
    print "\nGenerating environment:\n\n";

    local @ARGV = $Glob->{'LCVer'}==0 ? ("-release", uc($Glob->{'APORel'}), "-own", $Glob->{'Own'}) : ("-release", uc($Glob->{'APORel'}), "-own", $Glob->{'Own'}, "-lcversion", "$Glob->{'LCVer'}");
    do "$ENV{TOOL}/bin/vmakeEnv.pl";
    if ( $@ )
    {
        print "\nInitial vmakeEnv failed.\n$@\n";
        exit (10);
    }

    if ($^O =~ /MSWin32/i)
    {
        open ENV_OUT, ">$Glob->{'Own'}\\LCAEnv.bat" or die "Can't open $Glob->{'Own'}\\LCAEnv.bat\n$!\n";
        print ENV_OUT "\@echo off\n";
        print ENV_OUT "set OWN=$Glob->{'Own'}\n";
        print ENV_OUT "set P4USER=$Glob->{'P4User'}\n";
        print ENV_OUT "set P4CLIENT=$Glob->{'P4Client'}\n";
        print ENV_OUT "set P4PORT=$Glob->{'P4Port'}\n";
        print ENV_OUT "set BRANCH_INFO=$Glob->{'APORel'}\n";
        print ENV_OUT "set DEPOT=$APOSettings::Depots{$Glob->{'APORel'}}/$Glob->{'APORel'}\n";
        print ENV_OUT "set FORCED_LCVERSION=$Glob->{'LCVer'}\n" if ($Glob->{'LCVer'} != 0);
        print ENV_OUT "set RELSTAT=PROD\n" if ($Glob->{'RELSTAT'} == 1);
        print ENV_OUT "call %OWN%\\sys\\src\\iprofile.bat\n";
        close ENV_OUT or die "Can't close $Glob->{'Own'}\\LCAEnv.bat\n$!\n";

        open ENV_IN, "<$Glob->{'Own'}\\LCAEnv.bat" or die "Can't open $Glob->{'Own'}\\LCAEnv.bat for reading!\n$!\n";
        while (<ENV_IN>) { print "$_"; }
        close ENV_IN or die "Can't close $Glob->{'Own'}\\LCAEnv.bat!\n$!\n";
        if ($Opts{_dosbatch})
        {
            copy("$Glob->{'Own'}\\LCAEnv.bat", "$ENV{'LCAENV_TMP'}\\LCAEnv_TMP.bat");
        }
        else
        {
            print "\nOK!\n\n$Glob->{'Own'}\\LCAEnv.bat has been generated. To generate the correct\nenvironment, please execute this script.\n\n";
        }
    }
    else
    {
        {
            open ENV_OUT, ">tmpout.dat" or die "Can't open 'tmpout.dat' for writing!\n$!\n";
            print ENV_OUT "P4USER=$Glob->{'P4User'} export P4USER\n";
            print ENV_OUT "P4CLIENT=$Glob->{'P4Client'} export P4CLIENT\n";
            print ENV_OUT "P4PORT=$Glob->{'P4Port'} export P4PORT\n";
            print ENV_OUT "RELSTAT=PROD export RELSTAT\n" if ($Glob->{'RELSTAT'} == 1);
            close ENV_OUT or die "Can't close tmpout.dat!\n$!\n";

            my $platform = vmakeEnv::getPlatform();

            vmakeEnv::printInfo("-platform", $platform, "-apo_src", $Glob->{'Own'});
            vmakeEnv::createProfile ("-all", "-append", "$ENV{TOOL}/profiles/APODev/iprofile.$platform", "-append", "tmpout.dat");

            unlink ("